
#ifndef __PLUGIN__
#define __PLUGIN__
#include "PassManger.h"
#include <unordered_map>
#include <unordered_set>
#include <vector>
#include <string>

class ElimExtraJump : PassPlugin {
private:
    std::unordered_map<BasicBlock*, std::vector<CondBrInstruction*>*> BlockToCondMap;
    std::unordered_map<BasicBlock*, std::vector<UncondBrInstruction*>*> BlockToUncondMap;
    // FIX: Only jumpblock is not enough.
    std::unordered_map<BasicBlock*, BasicBlock*> ExtraToJump;
    std::vector<BasicBlock* > deleteblocks;

    bool isExtrBlock(BasicBlock* );
    void collectAllCondAndUncond(BasicBlock* );
    void replaceExtrBlock(BasicBlock* );
    void ElimAllExtrBlock();
public:
    ElimExtraJump(Unit* _unit) : PassPlugin(_unit){};
    void execute();
    void execute(Function* f);
};

class ConSameStructElim : PassPlugin {
private:
    std::vector<Instruction*> instrflow;
    std::set<Operand*> allParams;
    bool sameInstrFlowFlag = false;
    bool isFirstOperandSet = false;
    Operand* firstOperand = nullptr;
    int ParamNum = 100;

    std::vector<std::string> getParamInstructionFlow(Function*f, Operand *op);
    void getAllParamsFunction(Function* f);
    bool judgeSomeInstructionFlow(Function*f, Operand* param, std::vector<std::string>& firstparamvec);
    void mergeParamsInstruction(Function* f);
    std::vector<std::string> sortInstructionFlow(Function* f);
    void setFlag(bool flag) { 
        sameInstrFlowFlag = flag;
    }
    bool getFlag(){
        return sameInstrFlowFlag;
    }
    void setFirstOperandSet(bool flag) { 
        isFirstOperandSet = flag;
    }
    bool getFirstOperandSet(){
        return isFirstOperandSet;
    }
public: 
    ConSameStructElim(Unit* _unit) : PassPlugin(_unit){};
    void execute();
    void execute(Function* f);
};

struct PhiStruct {
    Operand* phiDef;
    std::unordered_map<BasicBlock*, Operand*> phisrcs;
    PhiStruct(PhiInstruction* phi) {
        phiDef = phi->getDef();
        for(auto it: phi->getSrcs()){
            phisrcs[it.first] = it.second;
        }
    }
    bool operator==(PhiStruct& phistruct) const{
        if(phisrcs.size() != phistruct.phisrcs.size()){
            return false;
        }
        for(auto it: phisrcs){
            if(it.second != phistruct.phisrcs[it.first]){
                return false;
            }
        }
        return true;
    }
};

class ElimExtrPhi : PassPlugin {
private:
    // 存储所有的phi指令, 方便后续进行phi指令的替换
    std::unordered_map<BasicBlock*, std::vector<PhiInstruction*>*> PhiInstList;
    // 存储所有的phi指令的替换逻辑
    std::unordered_map<Operand*, Operand*> ExtrPhiToOnlyPhi;
    // 存储现有的phi指令的结构
    std::vector<PhiStruct*> PhiStructList;
    // 存储当前进行预处理的phi指令
    std::vector<PhiInstruction*> PrePhiList;
    // 存储当前正在进行处理的块
    BasicBlock* phiBB;

    // 这一步只去消除没有用的phi指令的分支
    void processPhi(Function* f);
    // 这一步用来检测可以被替换的phi指令
    bool hasPhiReplaced(std::vector<PhiInstruction*> phivec);
    // 这一步用来替换所有的可以被替换的phi指令
    void updatePhiInstList(BasicBlock* bb);
    // 这一步用来删除无用phi指令和只有一个分支的phi指令
    void elimNoUsePhi(Function* f);
public:
    ElimExtrPhi(Unit* _unit) : PassPlugin(_unit){};
    void execute();
    void execute(Function* f);
};

#endif // __PLUGIN__