#include "Hoisting.h"

#define DEBUG_SWITCH_hs 0
#if DEBUG_SWITCH_hs
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

// https://web.archive.org/web/20170806083023/http://www.compileroptimizations.com/category/hoisting.htm

void Hoisting::execute() {
    detector = new LoopDetection(unit);
    detector->run();
    print("======== hoisting pass ========");
    for (auto it = detector->begin(); it != detector->end(); it++) {
        // if (!detector->isBaseLoop(*it))
        //     continue;
        // if ((*it)->size() > 1)
        //     continue;
        auto loop = *it;
        // do {
        invariant.clear();
        findDef(loop);
        findInvariants(loop);
        if (invariant.size() > 0)
            moveInvariantsOut(loop);
        // loop = detector->getParentLoop(loop);
        // } while (loop);
    }
    print("======== hoisting end ========");
}

void Hoisting::findDef(blockSet *loop) {
    for (auto bb : *loop) {
        auto head = bb->end();
        SymbolEntry *se;
        for (auto instr = head->getNext(); instr != head;
             instr = instr->getNext()) {
            if (instr->isStore()) {
                if (instr->getUse()[0]->getDef() &&
                    instr->getUse()[0]->getDef()->getParent() != bb)
                    continue;
                // Fix: 这有问题啊.........
                se = instr->getOpse()[0];
                if (se == nullptr)
                    continue;
                print("def-store: " << se->toStr());
                defInLoop.insert(se);
                // std::cout<<"def-opd: "<<se->toStr()<<"\n";
            } else if (instr->isPhi() && instr->getDst()->usersNum() > 0) {
                se = instr->getOpse()[0];
                if (se == nullptr)
                    continue;
                print("def-phi: " << se->toStr());
                defInLoop.insert(se);
                // std::cout<<"def-phi-opd: "<<se->toStr()<<"\n";
            }
        }
    }
}

bool Hoisting::FunctionCheck(CallInstruction *call) {
    IdentifierSymbolEntry *funcSe =
        dynamic_cast<IdentifierSymbolEntry *>(call->getSymEntry());
    print(funcSe->getName());
    if (funcSe->isSysy() || funcSe->getName() == "llvm.memset.p0.i32")
        return false;

    Function *func = dynamic_cast<Function *>(funcSe->getFunction());
    for (auto bb : func->getBlockList()) {
        for (auto instr = bb->begin(); instr != bb->end();
             instr = instr->getNext()) {
            if (instr->isStore()) {
                auto src = instr->getDef();
                // std::cout << src->toStr() << "\n";
                if (src->isGlobal()) {
                    // WARN 这里能找到吗？
                    if (instr->getUse()[0]->isParam())
                        continue;
                    return false;
                }
                if (src->getDef()->isGep()) {
                    auto gep = dynamic_cast<GepInstruction *>(src->getDef());
                    if (dynamic_cast<IdentifierSymbolEntry *>(gep->getArrSe())
                            ->isParam())
                        print("FunctionCheck 0");
                    return false;
                }
            } else if (instr->isCall()) {
                if (dynamic_cast<CallInstruction *>(instr)->getSymEntry() ==
                    funcSe)
                    return false;
                if (!FunctionCheck((CallInstruction *)instr))
                    return false;
            }
        }
    }
    print("FunctionCheck 1");
    return true;
}

// 寻找不变量
void Hoisting::findInvariants(blockSet *loop) {
    std::unordered_set<Instruction *> allInstructions;

    // 遍历循环中的所有基本块，把所有的指令加入到集合中
    for (auto bb : *loop) {
        auto head = bb->end();
        for (auto instr = head->getNext(); instr != head;
             instr = instr->getNext()) {
            for (auto opd : instr->getUse()) {
                print(opd->toStr() << " " << opd->isInvariant());
                if (opd->isConst()) {
                    print("1");
                    opd->setInvariant(1);
                    // print("const: " << opd->toStr())
                } else if (opd->isGlobal() &&
                           defInLoop.find(opd->getSymbolEntry()) !=
                               defInLoop.end()) {
                    print("2");
                    continue;
                } else if (opd->getOpdSe() &&
                           defInLoop.find(opd->getOpdSe()) == defInLoop.end()) {
                    print("3");
                    if (opd->getOpdSe()->getType()->isArray()
                        // || dynamic_cast<IdentifierSymbolEntry
                        // *>(opd->getOpdSe())
                        //         ->isGlobal()
                    )
                        continue; // todo: hosting数组/~全局变量~

                    // std::cout << "opd se: " << opd->getOpdSe() << " "
                    //           << opd->getOpdSe()->toStr() << "\n ";
                    opd->setInvariant(1);
                } else {
                    opd->setInvariant(0);
                }
            }
            allInstructions.insert(instr);
        }
    }

    bool change = false;
    // 持续对循环中的所有基本块进行处理，直到没有新的循环不变量生成
    do {
        change = false;
        // 对每个基本块进行处理
        for (auto bb : *loop) {
            // 对每个基本块中的指令进行处理
            auto head = bb->end();
            for (auto instr = head->getNext(); instr != head;
                 instr = instr->getNext()) {
                // print("======= " << instr->getInstName() << " START =======");

                // 跳过一些特定类型的指令以及在循环中没有定义的指令
                if (instr->isCmp() || instr->isUncond() || instr->isCond() ||
                    // instr->isCall() ||
                    instr->isRet() || instr->isAlloc() || instr->isFtos() ||
                    instr->isStof() || instr->isBitcast() ||
                    instr->isGep() || // todo: hosting数组
                    allInstructions.find(instr) == allInstructions.end()) {
                    continue;
                }
                // if (instr->getDef()&&instr->getDef()->usersNum() < 1) {
                //     bb->remove(instr);
                //     continue;
                // }
                // 检查该指令的所有操作数是否都是循环不变量
                bool invar = true;
                if (instr->isCall()) {
                    CallInstruction *call =
                        dynamic_cast<CallInstruction *>(instr);
                    invar = FunctionCheck(call);
                    std::vector<Operand *> paramlist = call->getParams();
                    for (auto param : paramlist) {
                        if (!param->isInvariant()) {
                            invar = false;
                            break;
                        }
                    }
                } else {
                    bool cmpSrc = 0;
                    // 全局变量binary常量的cond
                    for (auto ins : instr->getDef()->getUse()) {
                        if (ins->isCmp()) {
                            cmpSrc = 1;
                            break;
                        }
                    }
                    if (cmpSrc)
                        continue;

                    for (auto opd : instr->getUse()) {
                        // std::cout<<"op = "<<opd->toStr()<<"\n";
                        print(instr->getInstName()
                              << " invar check: " << opd->toStr()
                              << " bool: " << opd->isInvariant());
                        if (!opd->isInvariant()) {
                            invar = false;
                            break;
                        }
                    }
                    if (instr->isStore()) {
                        if (!instr->getDef()->isInvariant())
                            invar = false;
                    }
                }
                if (invar) {
                    allInstructions.erase(instr);
                    invariant.push_back(std::make_pair(bb, instr));
                    if (instr->isStore())
                        instr->getUse()[0]->setInvariant(1);
                    else {
                        instr->getDef()->setInvariant(1);
                    }
                    change = true;
                }
            }
        }
    } while (change);

    bool allLoad = 1;
    print("============ LoopInvariants =============");
    for (auto pair : invariant) {
        if (!pair.second->isLoad())
            allLoad = 0;
        print("in bb: " << pair.first->getNo()
                        << "\tVal: " << (pair.second)->getInstName());
    }
    if (allLoad)
        invariant.clear();
}

// 外提不变量
void Hoisting::moveInvariantsOut(blockSet *loop) {
    bool allphi = 1;
    for (auto pair : invariant) {
        if (!pair.second->isPhi()) 
            allphi=0;
    }  
    if(allphi){
        invariant.clear();
        return;
    }

    BasicBlock *loopBB = *(loop->begin());
    BasicBlock *condBB = detector->getCondbb(loopBB);
    BasicBlock *endBB = detector->getEndbb(loopBB);

    BasicBlock *invarBB = new BasicBlock(condBB->getParent());
    print("== loopbb = " << loopBB->getNo() << " condbb = " << condBB->getNo()
                         << " endbb = " << endBB->getNo());
    print("  == new bb = " << invarBB->getNo());

    loopBB->removePred(condBB);
    loopBB->addPred(invarBB);
    condBB->removeSucc(loopBB);
    condBB->addSucc(invarBB);
    invarBB->addSucc(loopBB);
    invarBB->addPred(condBB);

    auto head_cond = condBB->end();
    for (auto instr = head_cond->getNext(); instr != head_cond;
         instr = instr->getNext()) {
        if (instr->isCond()) {
            dynamic_cast<CondBrInstruction *>(instr)->setTrueBranch(invarBB);
            break;
        } else if (instr->isUncond()) {
            dynamic_cast<UncondBrInstruction *>(instr)->setBranch(invarBB);
            break;
        }
    }
    for (auto pair : invariant) {
        Instruction *instr = pair.second;
        // print("remove: "<<instr->getInstName());
        pair.first->remove(instr);
        if (instr->isPhi()) {
            invarBB->insertFront(instr);
        } else
            invarBB->insertBack(instr);
    }

    new UncondBrInstruction(loopBB, invarBB);
    // new UncondBrInstruction(invarBB, condBB);

    // std::vector<BasicBlock *> labels;
    // std::vector<Operand *> vals;
    // std::vector<BasicBlock *> inloopLabels;
    // std::vector<Operand *> inloopVals;
    auto head = invarBB->end();
    // 处理phi指令
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (!instr->isPhi())
            continue;
        PhiInstruction *phi = dynamic_cast<PhiInstruction *>(instr);
        // labels.clear();
        // vals.clear();
        // inloopVals.clear();
        // inloopLabels.clear();
        // auto phiBBs = phi->getPhiBBs();
        // auto phiOperands = phi->getPhiOperands();

        auto bblist = invarBB->getParent()->getBlockList();
        std::set<BasicBlock *> removebbs;
        for (auto src : phi->getSrcs()) {
            // print("PHIBB " << src.first->getNo() << " " << src.second->toStr());
            if (std::find(bblist.begin(), bblist.end(), src.first) ==
                bblist.end()) {
                removebbs.insert(src.first);
            }
        }
        for (auto b : removebbs) {
            print("remove bb = " << b->getNo() << " "
                                 << phi->getSrc(b)->toStr());
            phi->getSrcs().erase(b);
        }
        removebbs.clear();

        // 就应该只有两个
        BasicBlock *inloop = nullptr;
        BasicBlock *outloop = nullptr;
        std::map<BasicBlock *, Operand *> in_loop_srcs;
        std::map<BasicBlock *, Operand *> out_loop_srcs;
        for (auto src : phi->getSrcs()) {
            if (loop->find(dynamic_cast<BasicBlock *>(src.first)) !=
                loop->end()) {
                inloop = src.first;
            } else {
                outloop = src.first;
            }
        }
        if (inloop != nullptr&&phi->getSrcs()[inloop]) {
            auto val = phi->getSrcs()[inloop];
            // print("invar src: " << outloop->getNo() << " " << val->toStr());
            phi->getSrcs()[outloop] = val;
            phi->getSrcs().erase(inloop);
        }
        // for (unsigned i = 0; i < 2; i++) {
        //     if (loop->find(dynamic_cast<BasicBlock *>(phiBBs[i])) !=
        //         loop->end()) {
        //         inloopLabels.push_back(phiBBs[i]);
        //         inloopVals.push_back(phiOperands[i]);
        //     } else {
        //         labels.push_back(phiBBs[i]);
        //         vals.push_back(phiOperands[i]);
        //         phiOperands[i]->removeUse(phi);
        //     }
        // }

        // if (labels.size() == 0)
        //     continue;
        // if (labels.size() == 1) {
        //     for (unsigned i = 0; i < instr->getNumOperand(); i += 2) {
        //         if (vals[0] != v[i])
        //             continue;
        //         instr->setOperand(i + 1, invarBB);
        //         labels[0]->removeUse(instr, i + 1);
        //     }
        //     continue;
        // }
        // else {

        //     for (int j = 0; j < 2; j++) {
        //         if (phi->getPhiBBs()[j] == labels[0]) {
        //             print(phi->getPhiBBs()[j]->getNo()
        //                   << " " << labels[0]->getNo());

        //             phi->getPhiOperands_v()[j] = inloopVals[0];

        //             phi->getSrcs()[labels[0]] = inloopVals[0];

        //             print(phi->getPhiOperands_v()[j]->toStr()
        //                   << " " << inloopVals[0]->toStr());

        //             phi->getPhiOperands_v()[j] = inloopVals[0];
        //         }
        //     }
        // }
        // auto new_phi = PhiInst::createPhi(v[0]->getType(), invarBB);
        // // assert(new_phi->getNumOperand() == 0);
        // for (unsigned i = 0; i < labels.size(); i++) {
        //     new_phi->addOperand(vals[i]);
        //     new_phi->addOperand(labels[i]);
        // }
        // instr->clearOps();
        // for (unsigned i = 0; i < inloopLabels.size(); i++) {
        //     instr->addOperand(inloopVals[i]);
        //     instr->addOperand(inloopLabels[i]);
        // }
        // instr->addOperand(new_phi);
        // instr->addOperand(invarBB);
        // invarBB->addInstrBegin(new_phi);
    }
    auto head_loop = loopBB->end();
    for (auto instr = head_loop->getNext(); instr != head_loop;
         instr = instr->getNext()) {
        if (!instr->isPhi())
            continue;
        PhiInstruction *phi = dynamic_cast<PhiInstruction *>(instr);
        if (phi->getSrcs().find(condBB) != phi->getSrcs().end()) {
            auto val = phi->getSrcs()[condBB];
            print("loop src: " << condBB->getNo() << " " << val->toStr());
            phi->getSrcs()[invarBB] = val;
            phi->getSrcs().erase(condBB);
        }
        // auto phiBBs = phi->getPhiBBs();
        // auto phiOperands = phi->getPhiOperands();

        // for (int j = 0; j < 2; j++) {
        //     if (phiBBs[j] != loopBB) {

        //         phi->getPhiBBs()[j] = invarBB;
        //         phi->getSrcs()[invarBB] = phi->getSrcs()[phiBBs[j]];
        //         phi->getSrcs().erase(phiBBs[j]);
        //     }
        // }
        // 删掉原有bb块
        // phi->getSrcs().erase(loopBB);
    }
    std::cout << "yes hoist done.\n";
}