#include "LoopDetection.h"

using blockSet = std::unordered_set<BasicBlock *>;

#define DEBUG_SWITCH_ld 0
#if DEBUG_SWITCH_ld
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif


void LoopDetection::clear() {
    loops.clear();
    coreBBs.clear();
}

void LoopDetection::printloop(blockSet *loop) {
    for (auto bb : *loop) {
        print(" | bb:" << bb->getNo());
    }
}

BasicBlock *LoopDetection::getCondbb(BasicBlock *loopbb) {
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto pred : loopbb->getPredBlock()) {
        // print("== pred: "<<pred->getNo());
        if (pred != loopbb)
            return pred;
    }
}

BasicBlock *LoopDetection::getOutbb(BasicBlock *condbb) {
    auto head = condbb->end();
    CondBrInstruction *cond;
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (instr->isCond()) {
            cond = dynamic_cast<CondBrInstruction *>(instr);
            break;
        }
    }
    return cond->getFalseBranch();
}

BasicBlock *LoopDetection::getEndbb(BasicBlock *loopbb) {
    // if((loopbb->gketSuccBlock()).size()!=1)assert(0);
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto succ : loopbb->getSuccBlock()) {
        // print("== succ: "<<succ->getNo());
        if (succ != loopbb)
            return succ;
    }
}

void test(Unit *unit) {
    auto detector = new LoopDetection(unit);
    print("start run");

    detector->run();
    print("run finish");
    for (auto it = detector->begin(); it != detector->end(); it++) {
        // if (!detector->isBaseLoop(*it))
        //     continue;
        auto loop = *it;
        print("size = " << loop->size());
        detector->printloop(loop);
        BasicBlock *loopbb;
        BasicBlock *condbb;
        BasicBlock *endbb;
        // if (loop->size() == 1) {
        //     BasicBlock *loopbb = *(loop->begin());
        //     condbb = detector->getCondbb(loopbb);
        //     endbb = detector->getEndbb(loopbb);
        // } else {
        //     condbb = detector->getLoopCond(loop);
        //     endbb = detector->getOutbb(condbb);
        // }
        print("-- condbb = " << condbb->getNo()
                             << " endbb = " << endbb->getNo());
    }
}

void LoopDetection::run() {

    for (auto func : unit->get_functions()) {
        for (auto bb : func->getBlockList()) {
            for (Instruction *ins = bb->begin(); ins != bb->end();
                 ins = ins->getNext()) {
                if (ins->isCond() &&
                    dynamic_cast<CondBrInstruction *>(ins)->getTrueBranch() ==
                        bb) {
                    // std::cout<<"??????? " << bb->getNo()<<"\n";
                    auto bset = new blockSet;
                    bset->insert(bb);
                    loops.insert(bset);
                }
            }
        }
    }
}


LoopDetection::~LoopDetection() {
    for (auto set : loops) {
        if (set)
            delete set;
    }
}
