#include "Unit.h"
#include "Ast.h"
#include "SymbolTable.h"
#include "Type.h"
#include <algorithm>
#include <stack>
#include <string>
extern FILE *yyout;

void Unit::insertFunc(Function *f) { func_list.push_back(f); }

void Unit::removeFunc(Function *func) {
    func_list.erase(std::find(func_list.begin(), func_list.end(), func));
}

void Unit::removeGlobal(SymbolEntry *se) { 
    if(std::find(global_list.begin(),global_list.end(),se)!=global_list.end()){
        global_list.erase(std::find(global_list.begin(),global_list.end(),se));
    }
 }

void Unit::removeDeclare(SymbolEntry *se) { 
    if(std::find(declare_list.begin(),declare_list.end(),se)!=declare_list.end()){
        declare_list.erase(std::find(declare_list.begin(),declare_list.end(),se));
    }
 }

void Unit::insertGlobal(SymbolEntry *se) { global_list.push_back(se); }

void Unit::insertDeclare(SymbolEntry *se) {
    auto it = std::find(declare_list.begin(), declare_list.end(), se);
    if (it == declare_list.end()) {
        declare_list.push_back(se);
    }
}

void Unit::output() const {
    for (auto se : global_list) {
        if (se->getType()->isInt()) {
            fprintf(yyout, "@%s = global %s %d, align 4\n", se->toStr().c_str(),
                    se->getType()->toStr().c_str(),
                    (int)((IdentifierSymbolEntry *)se)->getValue());

        } else if (se->getType()->isFloat()) {
            double temp = (float)(((IdentifierSymbolEntry *)se)->getValue());
            uint64_t val = reinterpret_cast<uint64_t &>(temp);

            fprintf(yyout, "@%s = global %s 0x%lX, align 4\n",
                    se->toStr().c_str(), se->getType()->toStr().c_str(), val);
            // format floating point to llvm format.

        } else if (se->getType()->isArray()) {
            ArrayType *type = (ArrayType *)(se->getType());
            // int size = type->getSize() / TypeSystem::intType->getSize();
            double *val = ((IdentifierSymbolEntry *)se)->getArrayValue();
            int i = 0;
            fprintf(yyout, "@%s = global ", se->toStr().c_str());
            if (((IdentifierSymbolEntry *)se)->isAllZero()) {
                fprintf(yyout, "%s zeroinitializer", type->toStr().c_str());
            } else {
                std::vector<ArrayType *> stk;
                std::vector<int> stk1;
                stk.push_back(type);
                stk1.push_back(0);
                ArrayType *temp;
                while (!stk.empty()) {
                    temp = stk[stk.size() - 1];
                    Type *elemType = temp->getElementType();
                    if (elemType->isInt() || elemType->isFloat()) {
                        fprintf(yyout, "%s [", temp->toStr().c_str());
                        for (int j = 0; j < temp->getLength(); j++) {
                            if (j != 0)
                                fprintf(yyout, ", ");
                            if (elemType->isInt()) {
                                fprintf(yyout, "i32 %d", (int)val[i++]);
                            } else if (elemType->isFloat()) {
                                fprintf(yyout, "float %f", (float)val[i++]);
                            }
                        }
                        fprintf(yyout, "]");
                        stk1.pop_back();
                        stk.pop_back();
                        if (stk.empty())
                            break;
                        stk1[stk1.size() - 1]++;
                        continue;
                    }

                    if (stk1[stk1.size() - 1] != temp->getLength()) {
                        stk.push_back((ArrayType *)(temp->getElementType()));
                        if (stk1[stk1.size() - 1] == 0)
                            fprintf(yyout, "%s [", temp->toStr().c_str());
                        if (stk1[stk1.size() - 1] != 0)
                            fprintf(yyout, ", ");
                        stk1.push_back(0);
                    } else {
                        fprintf(yyout, "]");
                        stk.pop_back();
                        stk1.pop_back();
                        if (stk.empty())
                            break;
                        stk1[stk1.size() - 1]++;
                        continue;
                    }
                }
            }
            fprintf(yyout, ", align 4\n");
        }
    }
    for (auto &func : func_list) {
        // std::cout << ((IdentifierSymbolEntry*)func->getSymPtr())->getName()
        // << std::endl;
        func->output();
    }
    for (auto se : declare_list) {
        FunctionType *type = (FunctionType *)(se->getType());
        std::string str = type->toStr();
        std::string name = str.substr(0, str.find('('));
        std::string param = str.substr(str.find('('));
        fprintf(yyout, "declare %s %s%s\n", type->getRetType()->toStr().c_str(),
                se->toStr().c_str(), param.c_str());
    }
}

void Unit::genMachineCode(MachineUnit *munit) {
    AsmBuilder *builder = new AsmBuilder();
    builder->setUnit(munit);
    for (auto &func : func_list) {
        func->genMachineCode(builder);
    }
}

Unit::~Unit() {
    for (auto &func : func_list)
        delete func;
}
